/*
Quadrature decoder using PIO0 to decode PHASE_A and PHASE_B signals at GPIO10 and GPIO11.
A preload value can be set (X-reg of PIO0_SM0) and is loaded at '0' state of INDEX_PIN GPIO12.

Each rising / falling edge is counted up to 10e7 counts/s. Variable 'q_cnt0' is updated via
DMA-channel CH8. After max. transfer-count of CH8, DMA-CH9 is used to retrigger CH8.

2024-03-13
Michael Nowak
http://www.mino-elektronik.de
Alle Angaben ohne Gewaehr !
*/

#include "rp2040.h"

// inputs
#define PHASE_A       10
#define PHASE_B       11
#define INDEX_PIN     12

// unconditional JMP = INSTR_MEM_adr
#define JMP_DEC       23
#define JMP_LOOP      15
#define JMP_LOOPx     19
#define JMP_TEST      17
#define JMP_INC       25
#define JMP_INC_CONT  27
#define JMP_OUT       28
#define JMP_PRELOAD   29
#define JMP_OUTx      30

#define DMA_COUNT     0xFFFFFFFF

int q_cnt0, q_cnt0_offset;     
static uint32_t dma_dummy;

void init_PIO0(void)
{
  uint32_t *prg;
// enable PIO0 
  RESETS_CLR->RESET = RESETS_RESET_pio0_Msk;    
  while (!(RESETS->RESET_DONE & RESETS_RESET_pio0_Msk));
  
// enable GPIOs
  while((RESETS->RESET_DONE & (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk)) !=
        (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk));

// PhaseA: GPIO10, PhaseB: GPIO11, Index: GPIO12
  IO_BANK0->GPIO10_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio0_0;
  IO_BANK0->GPIO11_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio0_0;
  IO_BANK0->GPIO12_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio0_0;
  PADS_BANK0->GPIO10 = PADS_BANK0_GPIO0_IE_Msk |              // pullup + schmitttrigger
                      PADS_BANK0_GPIO0_PUE_Msk | 
                      PADS_BANK0_GPIO0_SCHMITT_Msk;
  PADS_BANK0->GPIO11 = PADS_BANK0_GPIO0_IE_Msk |              // pullup + schmitttrigger
                      PADS_BANK0_GPIO0_PUE_Msk | 
                      PADS_BANK0_GPIO0_SCHMITT_Msk;
  PADS_BANK0->GPIO12 = PADS_BANK0_GPIO0_IE_Msk |              // pullup + schmitttrigger
                      PADS_BANK0_GPIO0_PUE_Msk | 
                      PADS_BANK0_GPIO0_SCHMITT_Msk;
  
// set output group starting @GPIO10
  PIO0->SM0_PINCTRL =  PHASE_A << PIO0_SM0_PINCTRL_IN_BASE_Pos;       // 10 + 11
  PIO0->SM0_EXECCTRL = (31 << PIO0_SM0_EXECCTRL_WRAP_TOP_Pos) | 
                       (JMP_LOOP << PIO0_SM0_EXECCTRL_WRAP_BOTTOM_Pos) |
                       (INDEX_PIN  << PIO0_SM0_EXECCTRL_JMP_PIN_Pos); // index-pin: 12
  
 
      prg = (uint32_t *)&PIO0->INSTR_MEM0;    // pgm_start

// 'xx' is new state
// 00xx state
      *prg++ = JMP_LOOP;
      *prg++ = JMP_DEC;
      *prg++ = JMP_INC;
      *prg++ = JMP_LOOP;

// 01xx state
      *prg++ = JMP_INC;
      *prg++ = JMP_LOOP;
      *prg++ = JMP_LOOP;
      *prg++ = JMP_DEC;

// 10xx state
      *prg++ = JMP_DEC;
      *prg++ = JMP_LOOP;
      *prg++ = JMP_LOOP;
      *prg++ = JMP_INC;

// 11xx state
      *prg++ = JMP_LOOP;
      *prg++ = JMP_INC;
      *prg++ = JMP_DEC;
      
// ADR15:      
// LOOP: test reset if X != Y
      *prg++ = JMP_TEST | 0xa0;     // JMP X != Y
      *prg++ = JMP_LOOPx;           // X == Y, don't test again
// ADR17:
// TEST: Reset, if INDEX_PIN == 0     
      *prg++ = JMP_LOOPx | 0xc0;    // JMP PIN == 1, no reset
      *prg++ = JMP_PRELOAD;         // and output X
      
// ADR19:      
// LOOPx: normal loop without reset
// JMP-adr calculation
      *prg++ = 0x60c2;              // OUT ISR, 2     get old state to 'xx'
      *prg++ = 0x4002;              // IN PINS, 2     new state = 'xx' << 2 + 'ba'
// OSR temp-reg for ISR
      *prg++ = 0xa0e6;              // MOV OSR, ISR   save act. state 'xxba' 
      *prg++ = 0xa0a6;              // MOV PC, ISR    jump to function
      
// ADR23:
// DEC: decrement Y and output
      *prg++ = JMP_DEC+1 + 0x80;    // JMP Y--        and continue
      *prg++ = JMP_OUT;             // JMP OUT

// ADR25:
// INC: increment Y 
      *prg++ = 0xa04a;              // MOV Y, ~Y      invert
      *prg++ = JMP_INC_CONT | 0x80; // JMP Y--,       Y-- and dummy JMP to INC_CONT
// INC_CONT:
      *prg++ = 0xa04a;              // MOV Y, ~Y      invert again
      
// ADR28:
// OUT: output Y via RXFIFO if no reset
      *prg++ = JMP_OUTx | 0xc0;     // JMP PIN == 1, no reset
// ADR29:
// PRELOAD: output X if reset
      *prg++ = 0xa041;              // MOV Y, X     Y = X = preload value
// ADR30:
// OUTx: output Y via RXFIFO      
      *prg++ = 0x4040;              // MOV ISR, Y     copy to ISR
// ADR31: wrapping to LOOP
      *prg++ = 0x8000;              // PUSH noblock   send without blocking, recieve via DMA

      
  PIO0->SM0_SHIFTCTRL = 0 << PIO0_SM0_SHIFTCTRL_IN_SHIFTDIR_Pos | // IN shift left
                        1 << PIO0_SM0_SHIFTCTRL_OUT_SHIFTDIR_Pos; // OUT shift right
  PIO0->SM0_CLKDIV = 1 << PIO0_SM0_CLKDIV_INT_Pos;                // default, max. speed
  PIO0_SET->CTRL = (1 << (PIO0_CTRL_SM_ENABLE_Pos));              // enable PIO1-SM0
}

// if needed all interrupts should be disabled
void preload_qcnt0_val(int x)
{
  PIO0_CLR->CTRL = (1 << (PIO0_CTRL_SM_ENABLE_Pos));      // disable PIO0-SM0
  PIO0->TXF0 = x;                         // write new value
  PIO0->SM0_INSTR = 0x8080;               // PULL OSR
  PIO0->SM0_INSTR = 0xa027;               // MOV X, OSR
  PIO0->SM0_INSTR = 0xa041;               // MOV Y, X     Y = X = preload value
  PIO0_SET->CTRL = (1 << (PIO0_CTRL_SM_ENABLE_Pos));      // enable PIO0-SM0
}  


// no ISR needed: get 'q_cnt0' via DMA->CH8
// reload CH8_TRANS_COUNT via DMA->CH9
void init_qcnt0(int preload)
{
  RESETS_SET->RESET = RESETS_RESET_dma_Msk;
  RESETS_CLR->RESET = RESETS_RESET_dma_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_dma_Msk));

// regular transfer  
  DMA->CH8_READ_ADDR = (uint32_t) &PIO0->RXF0;
  DMA->CH8_WRITE_ADDR = (uint32_t) &q_cnt0;
  DMA->CH8_TRANS_COUNT = DMA_COUNT;
  DMA->CH8_CTRL_TRIG =  DREQ_PIO0_RX0 << DMA_CH0_CTRL_TRIG_TREQ_SEL_Pos | // SM0-RXF0
                        9 << DMA_CH0_CTRL_TRIG_CHAIN_TO_Pos |        // trigger CH9
                        2 << DMA_CH0_CTRL_TRIG_DATA_SIZE_Pos |        // word
                        DMA_CH0_CTRL_TRIG_EN_Msk;                     // enable
      
// retrigger CH8_TRANS_COUNT
  DMA->CH9_READ_ADDR =  (uint32_t) &dma_dummy;                       // dummy location
  DMA->CH9_WRITE_ADDR = (uint32_t) &dma_dummy;
  DMA->CH9_TRANS_COUNT = 1;                                          // only 1 reload-transfer
  DMA->CH9_CTRL_TRIG = DREQ_AUTO << DMA_CH0_CTRL_TRIG_TREQ_SEL_Pos | // free running
                        8 << DMA_CH0_CTRL_TRIG_CHAIN_TO_Pos |        // retrigger CH8  
                        2 << DMA_CH0_CTRL_TRIG_DATA_SIZE_Pos |        // word
                        DMA_CH0_CTRL_TRIG_EN_Msk;                     // enable
  init_PIO0();
  preload_qcnt0_val(preload);
}

// get raw data
int get_qcnt0(void)
{
  return(q_cnt0);
}

// read rel. data
int read_qcnt0(void)
{
  return(q_cnt0 - q_cnt0_offset);
}

// preset act. position to x
void preset_qcnt0(int x)
{
  q_cnt0_offset = -q_cnt0 + x;
}

// set active state of INDEX_PIN
void set_index0_pol(int i)
{
  if(i) {                                                       // invert INDEX_PIN polarity
    PADS_BANK0->GPIO12 = PADS_BANK0_GPIO0_IE_Msk |              // pulldown + schmitttrigger
                      PADS_BANK0_GPIO0_PDE_Msk | 
                      PADS_BANK0_GPIO0_SCHMITT_Msk;
    IO_BANK0->GPIO12_CTRL = IO_BANK0_GPIO0_CTRL_INOVER_INVERT << IO_BANK0_GPIO0_CTRL_INOVER_Pos |
                          IO_BANK0_GPIO0_CTRL_FUNCSEL_pio0_0;

  } else {                                                      // normal operation active '0'
    PADS_BANK0->GPIO12 = PADS_BANK0_GPIO0_IE_Msk |              // pullup + schmitttrigger
                      PADS_BANK0_GPIO0_PUE_Msk | 
                      PADS_BANK0_GPIO0_SCHMITT_Msk;
    IO_BANK0->GPIO12_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_pio0_0; 
  }                      
}